// Software License Agreement (BSD License)
//
// Copyright (c) 2008, Tully Foote
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above
//    copyright notice, this list of conditions and the following
//    disclaimer in the documentation and/or other materials provided
//    with the distribution.
//  * Neither the name of the Willow Garage nor the names of its
//    contributors may be used to endorse or promote products derived
//    from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
// FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
// COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
// INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
// BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
// ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.

#include "velodyneDriver.hh"

using namespace std;

// PUBLIC METHODS

// CONSTRUCTOR & DESTRUCTOR
VelodyneDriver::VelodyneDriver() {
  // Confirm that the UDP message buffer matches the structure size
  assert(sizeof(velodyne_packet_t) == VELODYNE_MSG_BUFFER_SIZE);

  setup_socket();
}

VelodyneDriver::~VelodyneDriver() {
  close(udp_rx_sock);
}

void VelodyneDriver::setup_socket() {
  // Setup a UDP socket to receive packets from the Velodyne
  udp_rx_sock = socket(PF_INET, SOCK_DGRAM, 0);
  if ( udp_rx_sock == -1 ) {
    std::cerr<<"Velodyne driver could not create socket!"<<std::endl;
    throw std::exception();
  }

  /* Code for setting and getting buffer size.  Defautl seems to be 100k which is fine
  int err;
  int    sockbufsize = 0;
  socklen_t    size = sizeof(int);
  err =  getsockopt(udp_rx_sock, SOL_SOCKET, SO_RCVBUF, (char *)&sockbufsize, &size);
  cout << "Socket_buff_length " << sockbufsize << endl;
  //  err = setsockopt(udp_rx_sock, SOL_SOCKET, SO_RCVBUF, (char *)&sndsize,
  //		   (int)sizeof(sndsize));
  //uint8_t val = 1;
  //setsockopt(udp_rx_sock, SOL_SOCKET, SO_REUSEADDR, &val, sizeof(uint8_t));
  */
  
  // Bind address to UDP socket
  memset((char*)&sensor_addr, 0, sizeof(sensor_addr));

  sensor_addr.sin_family = AF_INET;
  sensor_addr.sin_port = htons(VELODYNE_UDP_PORT);
  sensor_addr.sin_addr.s_addr = INADDR_ANY;
  
  int udp_rx_bind = bind(udp_rx_sock, (struct sockaddr*)&sensor_addr, sizeof(sensor_addr));

  if ( udp_rx_bind != 0) {
    std::cerr<<"Velodyne driver could not bind socket!"<<std::endl;
    throw std::exception();
  }
}

int VelodyneDriver::rx_udp_packet(velodyne_packet_t &packet, int recv_flags) { 
  addr_size = sizeof(sensor_addr); // addr_size cannot overflow when passed

  udp_rx_size = recvfrom( udp_rx_sock,
                          &packet, VELODYNE_MSG_BUFFER_SIZE,
                          recv_flags,
                          (struct sockaddr *) &sensor_addr, &addr_size );

#if 0
  if(udp_rx_size <= 0 && !(recv_flags & MSG_DONTWAIT)) {
    std::cerr<<"Velodyne driver could not receive UDP packet! "<<std::endl;
    std::cerr<<"error: "<<errno
      <<" rx_sock: "<<udp_rx_sock
      <<" packet: "<<&packet
      <<" recv_flags: "<<recv_flags
      <<" sockaddr: "<<&sensor_addr
      <<" addr_size: "<<addr_size<<std::endl;
  }
  //print_packet (packet);
#endif //0
    return udp_rx_size;
}

void VelodyneDriver::print_packet(velodyne_packet_t &packet) {
  
  for (int in = 0; in < 12 ; in++)
    {
      printf("Block %d, %d, %d:\n",in, packet.shots[in].lower_upper, packet.shots[in].rotational_pos);
      printf("Distance: ");
      for (int im = 0; im < 32 ; im ++)
	{
	  printf("%d\t", packet.shots[in].lasers[im].distance);
	}
      printf("\nIntensity: ");
      for (int im = 0; im < 32 ; im ++)
	{
	  printf("%d\t" , packet.shots[in].lasers[im].intensity);
	}
      printf ("\n");
    }
  printf("Status: \n");
  std::cout<< (double)packet.degrees + (double)packet.frac_degrees / 256.0  <<" " << packet.status_string<< std::endl;
  printf("\n\n\n\n");
  
}
