// Software License Agreement (BSD License)
//
// Copyright (c) 2008, Tully Foote
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above
//    copyright notice, this list of conditions and the following
//    disclaimer in the documentation and/or other materials provided
//    with the distribution.
//  * Neither the name of the Willow Garage nor the names of its
//    contributors may be used to endorse or promote products derived
//    from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
// FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
// COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
// INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
// BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
// ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.

#ifndef _VELODYNE_DRIVER_HH_
#define _VELODYNE_DRIVER_HH_

/* *************************************************************************
 *
 * Velodyne Ladar Header File
 * 
 * This is the header for the velodyne ladar interface drivers.   
 * 
 ************************************************************************* */

#include <iostream>
#include <time.h>
#include <stdio.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <errno.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>
#include <netinet/in.h>
#include <fstream>
#include <math.h>
#include <assert.h>
#include <stdint.h>
#include <exception>





class VelodyneDriver {
public:
  static const int VELODYNE_NUM_LASERS = 32; // The number of lasers per shot
  static const int VELODYNE_NUM_SHOTS = 12; // The number of shots per packet
  static const int MIN_SENSING_DISTANCE = 1000; //2m in 2mm units
  static const int MAX_SENSING_DISTANCE = 24000; //1200m in 2mm units
  static const int VELODYNE_MSG_BUFFER_SIZE = 1206; //The sides of a packet

  static const double VELODYNE_DRIVER_BROADCAST_FREQ_HZ = 50.0; //The rate of broadcast packets
  static const int VELODYNE_UDP_PORT = 2368; //The port the Velodyne sends to

  static const uint16_t VELODYNE_UPPER_HEADER_BYTES = 61183; //The byte indicating a upper shot
  static const uint16_t VELODYNE_LOWER_HEADER_BYTES = 56831; //The byte indicating a lower shot

  
  // Velodyne datastructures
  typedef struct vel_laser {
    uint16_t distance; // 2mm increments
    unsigned char intensity; // 255 being brightest
  } __attribute__((packed)) vel_laser_t;
  
  typedef struct velodyne_fire {
    uint16_t lower_upper;
    uint16_t rotational_pos; // 0-35999  divide by 100 for degrees
    vel_laser_t lasers[VELODYNE_NUM_LASERS]; 
  } __attribute__((packed)) velodyne_fire_t ;
  
  typedef struct velodyne_packet {
    velodyne_fire_t shots[VELODYNE_NUM_SHOTS];  
    uint8_t frac_degrees;
    uint8_t degrees;
    char status_string[4];  
  }  __attribute__((packed)) velodyne_packet_t ;
  
  
  // CONSTRUCTOR & DESTRUCTOR
  VelodyneDriver();
  ~VelodyneDriver();
 
  /* Get a new message
   * Return: socket read size
   * Arguments: A reference to a packet preallocated and receive flags.
   *
   * The receive_flags are passed through to the socket call,
   * possible values are defined in bits/socket.h Suggested MSG_NOWAIT
   */

  int rx_udp_packet(velodyne_packet_t& packet_reference,int socket_recieve_flags);  

  //Print the packet to the screen
  void print_packet(velodyne_packet_t &);

private:
  // File descriptor for the socket
  int udp_rx_sock;

  // Size of received data
  int udp_rx_size;

  // Size of address structure
  unsigned int addr_size;

  // Structure for storing the velodyne's port & ip address
  struct sockaddr_in sensor_addr;   
  
  // Message buffer
  unsigned char msgbuf[VELODYNE_MSG_BUFFER_SIZE];
  
  // Open and setup the socket to recieve
  void setup_socket(); 
};


#endif
